import os

from neuralogic.utils.data.dataset import Dataset
from neuralogic.core.builder import Backend
from neuralogic.core.settings import Settings

base_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "datasets")


def XOR(backend: Backend, settings: Settings):
    from neuralogic.core.problem import Problem

    return Problem.build_from_dir(os.path.join(base_path, "simple", "xor", "naive"), backend, settings)


def Trains(backend: Backend, settings: Settings):
    from neuralogic.core.problem import Problem

    return Problem.build_from_dir(os.path.join(base_path, "simple", "trains"), backend, settings)


def XOR_Vectorized(backend: Backend, settings: Settings):
    from neuralogic.core.problem import Problem

    return Problem.build_from_dir(os.path.join(base_path, "simple", "xor", "vectorized"), backend, settings)


def Mutagenesis(backend: Backend, settings: Settings):
    from neuralogic.core.problem import Problem

    return Problem.build_from_dir(os.path.join(base_path, "molecules", "mutagenesis"), backend, settings)



from neuralogic import get_neuralogic, get_gateway
from py4j.java_gateway import set_field, get_field
from enum import Enum


class Optimizer(str, Enum):
    ADAM = "ADAM"
    SGD = "SGD"


class Settings:
    def __init__(self, learning_rate: float = None, optimizer: Optimizer = None, epochs: int = None):
        self.namespace = get_neuralogic().cz.cvut.fel.ida.setup
        self.settings = self.namespace.Settings()

        if learning_rate is not None:
            self.learning_rate = learning_rate

        if optimizer is not None:
            self.optimizer = optimizer

        if epochs is not None:
            self.epochs = epochs

        set_field(self.settings, "debugExporting", False)
        set_field(self.settings, "isoValueCompression", False)
        set_field(self.settings, "exportBlocks", get_gateway().new_array(get_gateway().jvm.java.lang.String, 0))

        self.settings.infer()

    @property
    def seed(self) -> int:
        return get_field(self.settings, "seed")

    @seed.setter
    def seed(self, seed: int):
        set_field(self.settings, "seed", seed)

    @property
    def learning_rate(self) -> float:
        return get_field(self.settings, "initLearningRate")

    @learning_rate.setter
    def learning_rate(self, learning_rate: float):
        set_field(self.settings, "initLearningRate", learning_rate)

    @property
    def optimizer(self):
        return self.settings.getOptimizer()

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer):
        if optimizer == Optimizer.SGD:
            java_optimizer = self.namespace.Settings.OptimizerSet.SGD
        elif optimizer == Optimizer.ADAM:
            java_optimizer = self.namespace.Settings.OptimizerSet.ADAM
        else:
            raise NotImplementedError
        self.settings.setOptimizer(java_optimizer)

    @property
    def epochs(self) -> int:
        return get_field(self.settings, "maxCumEpochCount")

    @epochs.setter
    def epochs(self, epochs: int):
        set_field(self.settings, "maxCumEpochCount", epochs)

    @property
    def debug_exporting(self) -> bool:
        return get_field(self.settings, "debugExporting")

    @debug_exporting.setter
    def debug_exporting(self, debug_export: bool):
        set_field(self.settings, "debugExporting", debug_export)

    @property
    def default_fact_value(self) -> float:
        return get_field(self.settings, "defaultFactValue")

    @default_fact_value.setter
    def default_fact_value(self, value: float):
        set_field(self.settings, "defaultFactValue", value)

    def to_json(self) -> str:
        return self.settings.exportToJson()



import json
from collections import Sized
from typing import Optional, Dict

from neuralogic import get_neuralogic
from py4j.java_gateway import get_field

from neuralogic.core.settings import Settings


class Loss:
    def __init__(self, loss):
        self.loss = loss

    def backward(self):
        self.loss.backward()

    def value(self) -> float:
        return get_field(self.loss.getError(), "value")

    def output(self) -> float:
        return get_field(self.loss.getOutput(), "value")

    def target(self) -> float:
        return get_field(self.loss.getTarget(), "value")


class NeuraLogicLayer:
    def __init__(self, model, settings: Optional[Settings] = None):
        self.namespace = get_neuralogic().cz.cvut.fel.ida.neural.networks.computation.training.strategies
        self.do_train = True

        if settings is None:
            settings = Settings()

        self.settings = settings
        self.neural_model = model
        self.strategy = self.namespace.PythonTrainingStrategy(settings.settings, model)
        self.samples_len = 0
        self.reset_parameters()

    def reset_parameters(self):
        self.strategy.resetParameters()

    def train(self):
        self.do_train = True

    def test(self):
        self.do_train = False

    def set_training_samples(self, samples):
        self.samples_len = len(samples)
        self.strategy.setSamples(samples)

    def __call__(self, samples=None, train: bool = None, auto_backprop: bool = False, epochs: int = 1):
        if train is not None:
            self.do_train = train

        if samples is None:
            return self.strategy.learnSamples(epochs), self.samples_len

        if not isinstance(samples, Sized):
            if self.do_train:
                if auto_backprop:
                    return self.strategy.learnSample(samples), 1
            result = self.strategy.evaluateSample(samples)
            return Loss(result)

        if self.do_train:
            return self.strategy.learnSamples(samples, epochs), len(samples)

        results = self.strategy.evaluateSamples(samples)
        return [(get_field(result.getTarget(), "value"), get_field(result.getOutput(), "value")) for result in results]

    def state_dict(self) -> Dict:
        weights = self.neural_model.getAllWeights()
        weights_dict = {}

        for weight in weights:
            if get_field(weight, "isLearnable"):
                value = get_field(weight, "value")

                size = list(value.size())

                if len(size) == 0 or size[0] == 0:
                    weights_dict[get_field(weight, "index")] = get_field(value, "value")
                elif len(size) == 1 or size[0] == 1 or size[1] == 1:
                    weights_dict[get_field(weight, "index")] = list(get_field(value, "values"))
                else:
                    weights_dict[get_field(weight, "index")] = json.loads(value.toString())
        return {
            "weights": weights_dict,
        }

    def load_state_dict(self, state_dict: Dict):
        weights = self.neural_model.getAllWeights()
        weight_dict = state_dict["weights"]

        for weight in weights:
            if not get_field(weight, "isLearnable"):
                continue
            weight_value = get_field(weight, "value")

            index = get_field(weight, "index")
            value = weight_dict[index]

            if isinstance(value, (float, int)):
                weight_value.set(0, float(value))
                continue

            if isinstance(value[0], (float, int)):
                for i, val in enumerate(value):
                    weight_value.set(i, float(val))
                continue

            rows = len(value)

            for i, values in enumerate(value):
                for j, val in enumerate(values):
                    weight_value.set(i * rows + j, float(val))



class Dataset:
    def __init__(self, samples):
        self.__samples = samples
        self.__len = len(samples)

    def __len__(self):
        return self.__len

    def __getitem__(self, item):
        return self.__samples[item]

    @property
    def samples(self):
        return self.__samples



from typing import Iterable, Union
from neuralogic.core.constructs.predicate import Predicate
from neuralogic.core.constructs.java_objects import get_java_factory
from neuralogic.core.constructs import rule, factories


AtomType = Union["BaseAtom", "WeightedAtom"]
BodyAtomType = Union["BaseAtom", "WeightedAtom"]

Head = AtomType
Body = Union[Iterable[BodyAtomType], BodyAtomType]


class BaseAtom:
    def __init__(self, predicate: Predicate, terms=None, negated=False):
        self.predicate = predicate
        self.negated = negated
        self.terms = terms

        if self.terms is None:
            self.terms = []
        elif not isinstance(self.terms, Iterable):
            self.terms = [self.terms]

        self.java_object = get_java_factory().get_valued_fact(self, get_java_factory().get_variable_factory())

    def __neg__(self) -> "BaseAtom":
        return self.__invert__()

    def __invert__(self) -> "BaseAtom":
        return BaseAtom(self.predicate, self.terms, not self.negated)

    def __truediv__(self, other):
        if not isinstance(other, int) or self.predicate.arity != 0 or other < 0:
            raise NotImplementedError

        name, private, special = self.predicate.name, self.predicate.private, self.predicate.special
        return factories.AtomFactory.Predicate.get_predicate(name, other, private, special)

    def __call__(self, *args) -> "BaseAtom":
        if self.terms:
            raise Exception

        terms = list(args)
        arity = len(terms)

        name, private, special = self.predicate.name, self.predicate.private, self.predicate.special
        predicate = factories.AtomFactory.Predicate.get_predicate(name, arity, private, special)

        return BaseAtom(predicate, terms, self.negated)

    def __getitem__(self, item) -> "WeightedAtom":
        return WeightedAtom(self, item)

    def __le__(self, other: Body) -> rule.Rule:
        return rule.Rule(self, other)

    def to_str(self, end=False) -> str:
        negation = "~" if self.negated else ""
        end = "." if end else ""

        if self.terms:
            terms = ", ".join(str(term) for term in self.terms)
            return f"{negation}{self.predicate.to_str()}({terms}){end}"
        return f"{negation}{self.predicate.to_str()}{end}"

    def __str__(self) -> str:
        return self.to_str(True)

    def __copy__(self):
        atom = BaseAtom.__new__(BaseAtom)
        atom.negated = self.negated
        atom.terms = self.terms
        atom.predicate = self.predicate
        atom.java_object = self.java_object


class WeightedAtom:
    def __init__(self, atom: BaseAtom, weight, fixed=False):
        self.atom = atom
        self.weight = weight
        self.is_fixed = fixed

        if isinstance(weight, Iterable) and not isinstance(weight, tuple):
            self.weight = list(weight)

        self.java_object = get_java_factory().get_valued_fact(self, get_java_factory().get_variable_factory())

    def fixed(self) -> "WeightedAtom":
        if self.is_fixed:
            raise Exception
        return WeightedAtom(self.atom, self.weight, True)

    @property
    def negated(self):
        return self.atom.negated

    @property
    def predicate(self):
        return self.atom.predicate

    @property
    def terms(self):
        return self.atom.terms

    def __invert__(self) -> "WeightedAtom":
        return WeightedAtom(~self.atom, self.weight, self.is_fixed)

    def __neg__(self) -> "WeightedAtom":
        return self.__invert__()

    def __le__(self, other: Body) -> rule.Rule:
        return rule.Rule(self, other)

    def to_str(self, end=False):
        if isinstance(self.weight, tuple):
            weight = f"{{{', '.join(str(w) for w in self.weight)}}}"
        else:
            weight = str(self.weight)

        if self.is_fixed:
            return f"<{weight}> {self.atom.to_str(end)}"
        return f"{weight} {self.atom.to_str(end)}"

    def __str__(self):
        return self.to_str(True)

    def __copy__(self):
        atom = WeightedAtom.__new__(WeightedAtom)
        atom.atom = self.atom
        atom.weight = self.weight
        atom.is_fixed = self.is_fixed
        atom.java_object = self.java_object



from neuralogic.core.constructs.java_objects import get_java_factory
from neuralogic.core.constructs.metadata import Metadata
from typing import Iterable, Optional


class Rule:
    def __init__(self, head, body):
        self.head = head

        if not isinstance(body, Iterable):
            body = [body]

        self.body = list(body)
        self.metadata: Optional[Metadata] = None

        self.java_object = get_java_factory().get_rule(self)

    def __str__(self):
        metadata = "" if self.metadata is None is None else f" {self.metadata}"
        return f"{self.head.to_str()} :- {', '.join(atom.to_str() for atom in self.body)}.{metadata}"

    def __and__(self, other):
        if isinstance(other, Iterable):
            self.body.extend(list(other))
        else:
            self.body.append(other)
        return self

    def __or__(self, other):
        if not isinstance(other, Metadata):
            raise NotImplementedError
        self.metadata = other
        self.java_object.setMetadata(get_java_factory().get_metadata(self.metadata))

        return self



from typing import Optional, Tuple

from neuralogic.core.builder import Backend
from neuralogic.nn.base import AbstractEvaluator
from neuralogic.core import Problem
from neuralogic.core.settings import Settings
from neuralogic.utils.data import Dataset


def get_neuralogic_layer(backend: Backend):
    if backend == Backend.DYNET:
        from neuralogic.nn.dynet import NeuraLogicLayer  # type: ignore

        return NeuraLogicLayer
    if backend == Backend.DGL:
        from neuralogic.nn.dgl import NeuraLogicLayer  # type: ignore

        return NeuraLogicLayer
    if backend == Backend.JAVA:
        from neuralogic.nn.java import NeuraLogicLayer  # type: ignore

        return NeuraLogicLayer


def get_evaluator(
    backend: Backend,
    model_and_dataset=None,
    problem: Problem = None,
    settings: Optional[Settings] = None,
):
    model, dataset = None, None

    if problem is None and model_and_dataset is None:
        raise NotImplementedError
    if problem is not None and model_and_dataset is not None:
        raise NotImplementedError

    if settings is None:
        if problem is not None:
            settings = problem.java_factory.settings
        elif model_and_dataset is not None:
            model, dataset = model_and_dataset

            settings = model.settings
        else:
            raise Exception

    if backend == Backend.DYNET:
        from neuralogic.nn.evaluators.dynet import DynetEvaluator

        return DynetEvaluator(problem, model, dataset, settings)
    if backend == Backend.JAVA:
        from neuralogic.nn.evaluators.java import JavaEvaluator

        return JavaEvaluator(problem, model, dataset, settings)



from typing import List, Union, Tuple
from neuralogic.core.builder import Sample, Neuron
from neuralogic.error import MixedActivationFunctionsInLayerException


def unit_weight_generator():
    while True:
        yield -1


def to_layers(sample: Sample) -> List[List[Neuron]]:
    neuron_layers = [0] * len(sample.neurons)
    layers: List[List[Neuron]] = [[]]

    for neuron in sample.neurons:
        if len(neuron.inputs) == 0:
            layers[0].append(neuron)
            continue

        layer = max(neuron_layers[input] for input in neuron.inputs) + 1
        if len(layers) <= layer:
            layers.append([neuron])
            neuron_layers[neuron.index] = len(layers)
            continue

        for layer in layers[layer:]:
            if layer[0].activation == neuron.activation:
                layer.append(neuron)
                neuron_layers[neuron.index] = neuron_layers[layer[0].index]
                break
        else:
            layers.append([neuron])
            neuron_layers[neuron.index] = len(layers)
    return layers


class Layer:
    def __init__(self, layer: List[Neuron]):
        self.u = []
        self.v = []
        self.weights = []
        self.targets = []

        self.has_weights = False
        self.activation = -1
        self.value_shape: Union[Tuple[int], Tuple[int, int]] = (1,)

        for neuron in layer:
            if self.activation != -1 and neuron.activation != self.activation:
                raise MixedActivationFunctionsInLayerException()

            self.activation = neuron.activation
            self.targets.append(neuron.index)

            if neuron.value:
                if isinstance(neuron.value, list):
                    if isinstance(neuron.value[0], list):
                        self.value_shape = len(neuron.value), len(neuron.value[0])
                    else:
                        self.value_shape = (len(neuron.value),)
                else:
                    self.value_shape = (1,)

            weights = unit_weight_generator()
            if neuron.weights is not None and len(neuron.weights) != 0:
                weights = neuron.weights
                self.has_weights = True

            for input, weight in zip(neuron.inputs, weights):
                self.u.append(input)
                self.v.append(neuron.index)
                self.weights.append(weight)


class ProcessedSample:
    def __init__(self, sample: Sample):
        self.sample = sample
        self.layers = [Layer(layer) for layer in to_layers(sample)]

    @property
    def target(self):
        return float(self.sample.target)

    @property
    def output_neuron(self):
        return self.sample.output_neuron

    @property
    def neurons(self):
        return self.sample.neurons


def to_processed_samples(samples: List[Sample]) -> List[ProcessedSample]:
    return [ProcessedSample(sample) for sample in samples]



from typing import Union, Optional
from pathlib import Path
from contextlib import contextmanager
import os
from py4j.java_gateway import JavaGateway, JVMView

neuralogic: Optional[JVMView] = None
gateway: Optional[JavaGateway] = None

std_out = None
std_err = None


def set_java_home(java_home: Union[Path, str]):
    """
    Set java home environment variable

    :param java_home:
    :return:
    """
    os.environ["JAVA_HOME"] = str(java_home)


def set_std_out(out):
    global std_out
    std_out = out


def set_std_err(out):
    global std_err
    std_err = out


def initialize(die_on_exit=True):
    """
    Initialize the gateway for the Java process. Has to be reinitialized if some settings (JAVA_HOME) changes

    :param die_on_exit:
    :return:
    """
    global gateway, neuralogic

    if gateway is not None:
        try:
            gateway.shutdown()
        except Exception:
            pass

    gateway = JavaGateway.launch_gateway(
        classpath=os.environ["CLASSPATH"],
        redirect_stdout=std_out,
        redirect_stderr=std_err,
        die_on_exit=die_on_exit,
        daemonize_redirect=True,
    )

    neuralogic = gateway.jvm


def shutdown():
    if gateway is not None:
        gateway.shutdown()


def get_gateway() -> JavaGateway:
    """
    Get the gateway for the Java process

    :return:
    """
    if gateway is None:
        initialize()
    return gateway


def get_neuralogic() -> JVMView:
    """
    Get the jvm view of the Java process

    :return:
    """
    if neuralogic is None:
        initialize()
    return neuralogic


@contextmanager
def neuralogic_jvm():
    if gateway is None:
        initialize()
    yield
    if gateway is not None:
        shutdown()


os.environ["CLASSPATH"] = os.path.join(os.path.abspath(os.path.dirname(__file__)), "jar", "NeuraLogic.jar")



from neuralogic import get_neuralogic, get_gateway
from py4j.java_collections import ListConverter
from py4j.java_gateway import get_field, set_field

from typing import Union, List, Optional, Iterator, Tuple
from contextlib import contextmanager

from neuralogic.core.builder import Builder, Backend
from neuralogic.core.constructs.atom import BaseAtom, WeightedAtom
from neuralogic.core.constructs.rule import Rule
from neuralogic.core.constructs.predicate import PredicateMetadata
from neuralogic.core.constructs.java_objects import get_current_java_factory, set_java_factory, JavaFactory
from neuralogic.core.settings import Settings
from neuralogic.core.sources import Sources
from neuralogic.utils.data import Dataset

TemplateEntries = Union[BaseAtom, WeightedAtom, Rule]


def stream_to_list(stream) -> List:
    return list(stream.collect(get_gateway().jvm.java.util.stream.Collectors.toList()))


class Problem:
    def __init__(self, settings: Optional[Settings] = None):
        if settings is None:
            settings = Settings()

        self.java_factory = JavaFactory(settings)

        self.template: List[TemplateEntries] = []
        self.examples: List[TemplateEntries] = []
        self.queries: List[TemplateEntries] = []

        self.counter = 0

    def add_rule(self, rule):
        self.add_rules([rule])

    def add_rules(self, rules: List):
        self.template.extend(rules)

    def add_example(self, example):
        self.add_examples([example])

    def add_examples(self, examples: List):
        self.examples.extend(examples)

    def add_query(self, query):
        self.add_queries([query])

    def add_queries(self, queries: List):
        self.queries.extend(queries)

    def build_queries(self, query_builder):
        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.example
        logic_samples = []

        for query in self.queries:
            head, conjunction = self.java_factory.get_query(query)
            facts = get_field(conjunction, "facts")

            if head is not None:
                id = get_field(head, "literal").toString()
                if head.getValue() is None or not isinstance(query.head.weight, (float, int)):
                    logic_samples.extend(
                        [namespace.LogicSample(f.getValue(), query_builder.createQueryAtom(id, f), True) for f in facts]
                    )
                else:
                    importance = get_field(head.getValue(), "value")
                    logic_samples.extend(
                        [
                            namespace.LogicSample(f.getValue(), query_builder.createQueryAtom(id, importance, f), True)
                            for f in facts
                        ]
                    )
            else:
                id = str(self.counter)
                logic_samples.extend(
                    [namespace.LogicSample(f.getValue(), query_builder.createQueryAtom(id, f), True) for f in facts]
                )
            self.counter += 1
        return logic_samples

    def build_examples(self, examples_builder):
        logic_samples = []
        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.example

        for example in self.examples:
            label, lifted_example = self.java_factory.get_lifted_example(example)

            value = None
            label_fact = None if label is None else get_field(label, "facts")
            label_size = 0 if label is None else label_fact.size()

            if label is None or label_size == 0:
                query_atom = examples_builder.createQueryAtom(str(self.counter), None, lifted_example)
            elif label_size == 1:
                if label_fact.get(0).getValue() is None:
                    literal_string = get_field(label[0], "literal").toString()
                    query_atom = examples_builder.createQueryAtom(literal_string, label_fact.get(0), lifted_example)
                else:
                    value = label_fact.get(0).getValue()
                    query_atom = examples_builder.createQueryAtom(str(self.counter), label_fact.get(0), lifted_example)
            else:
                raise NotImplementedError
            logic_samples.append(namespace.LogicSample(value, query_atom))
            self.counter += 1
        return logic_samples

    def get_parsed_template(self):
        predicate_metadata = []
        weighted_rules = []
        valued_facts = []

        for rule in self.template:
            if isinstance(rule, PredicateMetadata):
                predicate_metadata.append(rule.java_object)
            elif isinstance(rule, Rule):
                weighted_rules.append(rule.java_object)
            elif isinstance(rule, (WeightedAtom, BaseAtom)):
                valued_facts.append(rule.java_object)

        weighted_rules = ListConverter().convert(weighted_rules, get_gateway()._gateway_client)
        valued_facts = ListConverter().convert(valued_facts, get_gateway()._gateway_client)
        predicate_metadata = ListConverter().convert(predicate_metadata, get_gateway()._gateway_client)

        template_namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.template.types
        return template_namespace.ParsedTemplate(weighted_rules, valued_facts)

    def merge_queries_with_examples(self, queries, examples):
        logic_samples = []

        if len(examples) == 0:
            return queries

        if len(queries) == 0:
            return examples

        if len(examples) == 1:
            for query in queries:
                example = examples[0] if get_field(get_field(query, "query"), "evidence") is None else query
                query_object = query if get_field(query, "isQueryOnly") else examples[0]

                example_evidence = get_field(get_field(example, "query"), "evidence")
                set_field(get_field(query_object, "query"), "evidence", example_evidence)
                logic_samples.append(query)
            return logic_samples

        if len(examples) != len(queries):
            raise Exception(f"The size of examples {len(examples)} doesn't match the size of queries {len(queries)}")

        for query, example in zip(queries, examples):
            example_object = example if get_field(get_field(query, "query"), "evidence") is None else query
            query_object = query if get_field(query, "isQueryOnly") else example

            example_evidence = get_field(get_field(example_object, "query"), "evidence")
            set_field(get_field(query_object, "query"), "evidence", example_evidence)
            logic_samples.append(query)
        return logic_samples

    def build(self, backend: Backend):
        from neuralogic.nn import get_neuralogic_layer

        self.counter = 0

        previous_factory = get_current_java_factory()
        set_java_factory(self.java_factory)

        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.building
        examples_builder = namespace.ExamplesBuilder(self.java_factory.settings.settings)
        query_builder = namespace.QueriesBuilder(self.java_factory.settings.settings)
        query_builder.setFactoriesFrom(examples_builder)

        examples = self.build_examples(examples_builder)
        queries = self.build_queries(query_builder)
        logic_samples = self.merge_queries_with_examples(queries, examples)

        logic_samples = ListConverter().convert(logic_samples, get_gateway()._gateway_client).stream()
        parsed_template = self.get_parsed_template()

        set_java_factory(previous_factory)
        weights, samples = Builder.from_problem(parsed_template, logic_samples, backend, self.java_factory.settings)

        return get_neuralogic_layer(backend)(weights, self.java_factory.settings), Dataset(samples)

    @contextmanager
    def context(self) -> Iterator["Problem"]:
        previous_factory = get_current_java_factory()
        set_java_factory(self.java_factory)
        yield self
        set_java_factory(previous_factory)

    def rules_to_str(self) -> str:
        return "\n".join(str(r) for r in self.template)

    def examples_to_str(self) -> str:
        return "\n".join(str(r) for r in self.examples)

    def queries_to_str(self) -> str:
        return "\n".join(str(r) for r in self.queries)

    @staticmethod
    def build_from_dir(directory: str, backend: Backend, settings: Settings, args: Optional[List] = None):
        from neuralogic.nn import get_neuralogic_layer

        args = [] if args is None else args

        args.extend(["-sd", str(directory)])
        sources = Sources.from_args(args, settings)

        weights, samples = Builder.from_sources(settings, backend, sources)

        return get_neuralogic_layer(backend)(weights, settings), Dataset(samples)

    @staticmethod
    def build_from_files(
        rules_file: str,
        backend: Backend,
        settings: Settings,
        example_file: Optional[str] = None,
        queries_file: Optional[str] = None,
        args: Optional[List] = None,
    ):
        from neuralogic.nn import get_neuralogic_layer

        args = [] if args is None else args

        args.extend(["-t", str(rules_file)])

        if queries_file is not None:
            args.extend(["-q", str(queries_file)])
        if example_file is not None:
            args.extend(["-e", str(example_file)])

        sources = Sources.from_args(args, settings)

        weights, samples = Builder.from_sources(settings, backend, sources)

        return get_neuralogic_layer(backend)(weights, settings), Dataset(samples)



from typing import Optional, Dict

from neuralogic.core.builder import Backend
from neuralogic.core import Problem
from neuralogic.nn.java import NeuraLogicLayer
from neuralogic.nn.base import AbstractEvaluator
from neuralogic.core.settings import Settings
from neuralogic.utils.data import Dataset


class JavaEvaluator(AbstractEvaluator):
    def __init__(
        self,
        problem: Optional[Problem],
        model: Optional[NeuraLogicLayer],
        dataset: Optional[Dataset],
        settings: Settings,
    ):
        super().__init__(problem, model, dataset, settings)

        if problem is not None:
            model, dataset = problem.build(Backend.JAVA)
            self.dataset = dataset

        self.neuralogic_layer: NeuraLogicLayer = model
        self.neuralogic_layer.set_training_samples(self.dataset.samples)

    def train(self, generator: bool = True, epochs: int = None):
        if epochs is None:
            epochs = self.settings.epochs

        def _train():
            for _ in range(epochs):
                result = self.neuralogic_layer(None, True)
                yield result

        if generator:
            return _train()

        return self.neuralogic_layer(None, True, epochs=epochs)

    def test(self, generator: bool = True):
        def _test():
            for sample in self.dataset.samples:
                result = self.neuralogic_layer(sample, False)
                yield result.target(), result.output()

        if generator:
            return _test()
        return self.neuralogic_layer(self.dataset.samples, False)

    def state_dict(self) -> Dict:
        return self.neuralogic_layer.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_layer.load_state_dict(state_dict)



import dgl
import dgl.function as fn
import torch
from typing import List
from torch.nn import Parameter, ParameterList, Sequential
from neuralogic.builder import Weight
from neuralogic.preprocessing import ProcessedSample, Layer


class NeuraLogicHelperLayer(torch.nn.Module):
    activations = {
        "Sigmoid": torch.sigmoid,
        "ReLu": torch.relu,
        "Tanh": torch.tanh,
        "Average": lambda x: torch.mean(x, 1, True),
        "Maximum": lambda x: torch.max(x, 1, True),
    }

    def __init__(self, layer: Layer, neuron_count: int, weights: ParameterList):
        super(NeuraLogicHelperLayer, self).__init__()

        self.reduce = "Sum" if layer.activation not in NeuraLogicHelperLayer.activations else layer.activation

        self.neuron_count = neuron_count
        self.layer = layer
        self.weights = weights

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if len(self.layer.u) == 0:
            return x

        # old_x = x.clone()
        # old_x[self.layer.targets] = 0

        g = dgl.graph((self.layer.u, self.layer.v), num_nodes=self.neuron_count)
        message = self.message if self.layer.has_weights else fn.copy_src(src="x", out="m")
        reduce = fn.sum(msg="m", out="x") if self.reduce == "Sum" else self.reduce_function

        with g.local_scope():
            g.ndata["x"] = x
            g.update_all(message, reduce)

            # if len(old_x.shape) < len(g.ndata["x"]):
            #     old_x = old_x.reshape(old_x.shape + tuple(1 for _ in range(len(g.ndata["x"].shape) - len(old_x.shape))))

            return g.ndata["x"]

    def reduce_function(self, nodes):
        x = torch.sum(nodes.mailbox["m"], dim=1)
        out = self.activations[self.reduce](x)

        return {"x": out}

    def message(self, edges):
        xs = edges.src["x"]
        out = [
            (self.weights[w] @ x) if x.shape != (1,) and self.weights[w].shape != (1,) else (x * self.weights[w])
            for x, w in zip(xs, self.layer.weights)
        ]

        return {"m": torch.stack(out)}


class NeuraLogicLayer(torch.nn.Module):
    def __init__(self, weights: List[Weight]):
        super(NeuraLogicLayer, self).__init__()

        params = [
            Parameter(
                torch.tensor(weight.value, dtype=torch.float).reshape(weight.dimensions), requires_grad=not weight.fixed
            )
            for weight in weights
        ]

        self.weights = ParameterList(params)
        self.meta_weights = weights
        self.reset_parameters()

    def reset_parameters(self):
        for weight, meta_weight in zip(self.weights, self.meta_weights):
            if not meta_weight.fixed:
                torch.nn.init.uniform_(weight)

    def forward(self, sample: ProcessedSample):
        neuron_count = len(sample.neurons)

        neural_layers = [NeuraLogicHelperLayer(layer, neuron_count, self.weights) for layer in sample.layers]
        model = Sequential(*neural_layers)

        x = torch.zeros((neuron_count, *sample.layers[0].value_shape))

        for neuron_index in sample.layers[0].targets:
            neuron = sample.neurons[neuron_index]
            x[neuron.index] = torch.tensor(neuron.value)

        x = model(x)
        return torch.take(x, torch.tensor([sample.neurons[-1].index]))



from neuralogic.core.constructs.java_objects import get_java_factory
from neuralogic.core.constructs.metadata import Metadata


class Predicate:
    """WeightedPredicate"""

    def __init__(self, name, arity, private, special):
        self.name = name
        self.arity = arity
        self.private = private
        self.special = special

        self.java_object = get_java_factory().get_predicate(self)

    def set_arity(self, arity):
        if self.arity == arity:
            return self

    def to_str(self):
        special = "@" if self.special else ""
        private = "*" if self.private else ""
        return f"{private}{special}{self.name}"

    def __str__(self):
        special = "@" if self.special else ""
        private = "*" if self.private else ""
        return f"{private}{special}{self.name}/{self.arity}"

    def __or__(self, other):
        if not isinstance(other, Metadata):
            raise NotImplementedError
        return PredicateMetadata(self, other)


class PredicateMetadata:
    def __init__(self, predicate: Predicate, metadata: Metadata):
        if metadata.aggregation is not None or metadata.learnable is not None:
            raise NotImplementedError

        self.predicate = predicate
        self.metadata = metadata

        self.java_object = get_java_factory().get_predicate_metadata_pair(self)

    def __str__(self):
        return f"{self.predicate} {self.metadata}"



from typing import Optional, Dict

import dynet as dy

from neuralogic.nn.dynet import NeuraLogicLayer
from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Problem
from neuralogic.core.settings import Settings, Optimizer
from neuralogic.core.builder import Backend
from neuralogic.utils.data import Dataset


class DynetEvaluator(AbstractEvaluator):
    trainers = {
        Optimizer.SGD: lambda param, rate: dy.SimpleSGDTrainer(param, learning_rate=rate),
        Optimizer.ADAM: lambda param, rate: dy.AdamTrainer(param, alpha=rate),
    }

    def __init__(
        self,
        problem: Optional[Problem],
        model: Optional[NeuraLogicLayer],
        dataset: Optional[Dataset],
        settings: Settings,
    ):
        super().__init__(problem, model, dataset, settings)

        if problem is not None:
            model, dataset = problem.build(Backend.DYNET)
            self.dataset = dataset
        self.neuralogic_layer = model

    def train(self, generator: bool = True):
        epochs = self.settings.epochs
        optimizer = Optimizer[str(self.settings.optimizer)]

        if optimizer not in DynetEvaluator.trainers:
            raise NotImplementedError

        trainer = DynetEvaluator.trainers[optimizer](self.neuralogic_layer.model, self.settings.learning_rate)

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0

                dy.renew_cg(immediate_compute=False, check_validity=False)

                for sample in self.dataset.samples:
                    label = dy.scalarInput(sample.target)
                    graph_output = self.neuralogic_layer(sample)
                    loss = dy.squared_distance(graph_output, label)

                    total_loss += loss.value()
                    loss.backward()
                    trainer.update()
                    seen_instances += 1

                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, generator: bool = True):
        def _test():
            for sample in self.dataset.samples:
                dy.renew_cg(immediate_compute=False, check_validity=False)

                graph_output = self.neuralogic_layer(sample)
                label = dy.scalarInput(sample.target)

                results = (label.value(), graph_output.value())
                yield results

        if generator:
            return _test()
        return list(_test())

    def state_dict(self) -> Dict:
        return self.neuralogic_layer.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_layer.load_state_dict(state_dict)



from neuralogic import get_neuralogic, get_gateway
from neuralogic.core.settings import Settings
from neuralogic.core.sources import Sources

import json
from enum import Enum
from typing import List, Optional
from py4j.java_gateway import get_field


def stream_to_list(stream) -> List:
    return list(stream.collect(get_gateway().jvm.java.util.stream.Collectors.toList()))


class Backend(Enum):
    DYNET = "dynet"
    PYG = "pyg"
    DGL = "dgl"
    JAVA = "java"


class Sample:
    def __init__(self, sample):
        self.id = get_field(sample, "id")
        self.target = float(get_field(sample, "target"))
        self.output_neuron = get_field(sample, "neuron")
        self.neurons = self.deserialize_network(get_field(sample, "network"))
        self.output_neuron = self.neurons[-1].index

    def deserialize_network(self, network):
        neurons = []

        for i, neuron in enumerate(network):
            neuron_object = Neuron(neuron, i)
            neurons.append(neuron_object)

            if neuron_object.index == self.output_neuron:
                break
        return neurons


class Neuron:
    def __init__(self, neuron, index):
        self.index = index
        self.name = get_field(neuron, "name")
        self.weighted = get_field(neuron, "weighted")
        self.activation = get_field(neuron, "activation")
        self.inputs = get_field(neuron, "inputs")
        self.weights = get_field(neuron, "weights")
        self.offset = get_field(neuron, "offset")
        self.value = get_field(neuron, "value")
        self.pooling = get_field(neuron, "pooling")

        if self.value:
            self.value = json.loads(self.value)

        if self.weights is not None:
            self.weights = list(self.weights)

        if self.inputs is not None:
            self.inputs = list(self.inputs)


class Weight(object):
    def __init__(self, weight):
        self.index: int = get_field(weight, "index")
        self.name = get_field(weight, "name")
        self.dimensions = tuple(get_field(weight, "dimensions"))
        self.value = json.loads(get_field(weight, "value"))
        self.fixed = get_field(weight, "isFixed")

        if not isinstance(self.value, list):
            self.value = self.value

        if not self.dimensions:
            self.dimensions = (1,)

    @staticmethod
    def get_unit_weight() -> "Weight":
        weight = Weight.__new__(Weight)
        weight.index = 0
        weight.name = "unit"
        weight.dimensions = (1,)
        weight.value = 1.0
        weight.fixed = True

        return weight


class Builder:
    @staticmethod
    def get_builders(settings: Settings, sources: Optional[Sources]):
        namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.building
        builder = namespace.End2endTrainigBuilder(settings.settings, None if sources is None else sources.sources)
        nn_builder = builder.getEnd2endNNBuilder()

        return builder, nn_builder

    @staticmethod
    def build(source_pipeline, sources: Optional[Sources]):
        pipes_namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.pipes.specific
        serializer_pipe = pipes_namespace.NeuralSerializerPipe()

        source_pipeline.connectAfter(serializer_pipe)
        source_pipeline.execute(None if sources is None else sources.sources)
        return serializer_pipe.get()

    @staticmethod
    def from_sources(settings: Settings, backend: Backend, sources: Sources):
        builder, nn_builder = Builder.get_builders(settings, sources)

        if backend == Backend.JAVA:
            source_pipeline = nn_builder.buildPipeline()
            source_pipeline.execute(None if sources is None else sources.sources)
            java_model = source_pipeline.get()

            logic_samples = get_field(java_model, "s")
            neural_model = get_field(java_model, "r")
            samples = logic_samples.collect(get_neuralogic().java.util.stream.Collectors.toList())
            return neural_model, samples

        result = Builder.build(nn_builder.buildPipeline(), sources)

        serialized_weights = list(get_field(result, "r"))
        weights: List = [None] * len(serialized_weights)

        for x in serialized_weights:
            weight = Weight(x)
            weights[weight.index] = weight

        samples = [Sample(x) for x in stream_to_list(get_field(result, "s"))]

        return weights, samples

    @staticmethod
    def from_problem(parsed_template, logic_samples, backend: Backend, settings: Settings):
        builder, nn_builder = Builder.get_builders(settings, None)

        if backend == Backend.JAVA:
            source_pipeline = nn_builder.buildPipeline(parsed_template, logic_samples)
            source_pipeline.execute(None)
            java_model = source_pipeline.get()

            logic_samples = get_field(java_model, "s")
            neural_model = get_field(java_model, "r")

            samples = logic_samples.collect(get_neuralogic().java.util.stream.Collectors.toList())
            return neural_model, samples

        result = Builder.build(nn_builder.buildPipeline(parsed_template, logic_samples), None)

        dummy_weight = Weight.get_unit_weight()
        serialized_weights = list(get_field(result, "r"))
        weights: List = [dummy_weight] * len(serialized_weights)

        for x in serialized_weights:
            weight = Weight(x)

            if weight.index >= len(weights):
                weights.extend([dummy_weight] * (weight.index - len(weights) + 1))
            weights[weight.index] = weight

        samples = [Sample(x) for x in stream_to_list(get_field(result, "s"))]

        return weights, samples



from neuralogic.grammar.NeuralogicListener import NeuralogicListener
from neuralogic.grammar.NeuralogicParser import NeuralogicParser
from neuralogic.grammar.NeuralogicVisitor import NeuralogicVisitor
from neuralogic.grammar.NeuralogicLexer import NeuralogicLexer



from typing import Optional

from neuralogic.core.settings import Settings
from neuralogic.core import Problem
from neuralogic.utils.data import Dataset


class AbstractNeuralogicLayer:
    def __call__(self, sample):
        raise NotImplementedError


class AbstractEvaluator:
    def __init__(self, problem: Optional[Problem], model: Optional, dataset: Optional[Dataset], settings: Settings):
        if model is None and problem is None:
            raise NotImplementedError
        if model is not None and problem is not None:
            raise NotImplementedError

        self.settings = settings
        self.dataset = dataset
        self.neuralogic_layer = model

    def train(self, generator: bool = True):
        pass

    def test(self, generator: bool = True):
        pass



# Generated from /home/lukas/Workspace/prcvut/pyneuralogic/neuralogic/grammar/Neuralogic.g4 by ANTLR 4.8
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .NeuralogicParser import NeuralogicParser
else:
    from NeuralogicParser import NeuralogicParser

# This class defines a complete generic visitor for a parse tree produced by NeuralogicParser.


class NeuralogicVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by NeuralogicParser#templateFile.
    def visitTemplateFile(self, ctx: NeuralogicParser.TemplateFileContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#templateLine.
    def visitTemplateLine(self, ctx: NeuralogicParser.TemplateLineContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#examplesFile.
    def visitExamplesFile(self, ctx: NeuralogicParser.ExamplesFileContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#liftedExample.
    def visitLiftedExample(self, ctx: NeuralogicParser.LiftedExampleContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#label.
    def visitLabel(self, ctx: NeuralogicParser.LabelContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#queriesFile.
    def visitQueriesFile(self, ctx: NeuralogicParser.QueriesFileContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#fact.
    def visitFact(self, ctx: NeuralogicParser.FactContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#atom.
    def visitAtom(self, ctx: NeuralogicParser.AtomContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#termList.
    def visitTermList(self, ctx: NeuralogicParser.TermListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#term.
    def visitTerm(self, ctx: NeuralogicParser.TermContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#variable.
    def visitVariable(self, ctx: NeuralogicParser.VariableContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#constant.
    def visitConstant(self, ctx: NeuralogicParser.ConstantContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#predicate.
    def visitPredicate(self, ctx: NeuralogicParser.PredicateContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#conjunction.
    def visitConjunction(self, ctx: NeuralogicParser.ConjunctionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#metadataVal.
    def visitMetadataVal(self, ctx: NeuralogicParser.MetadataValContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#metadataList.
    def visitMetadataList(self, ctx: NeuralogicParser.MetadataListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#lrnnRule.
    def visitLrnnRule(self, ctx: NeuralogicParser.LrnnRuleContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#predicateOffset.
    def visitPredicateOffset(self, ctx: NeuralogicParser.PredicateOffsetContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#predicateMetadata.
    def visitPredicateMetadata(self, ctx: NeuralogicParser.PredicateMetadataContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#weightMetadata.
    def visitWeightMetadata(self, ctx: NeuralogicParser.WeightMetadataContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#templateMetadata.
    def visitTemplateMetadata(self, ctx: NeuralogicParser.TemplateMetadataContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#weight.
    def visitWeight(self, ctx: NeuralogicParser.WeightContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#fixedValue.
    def visitFixedValue(self, ctx: NeuralogicParser.FixedValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#offset.
    def visitOffset(self, ctx: NeuralogicParser.OffsetContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#value.
    def visitValue(self, ctx: NeuralogicParser.ValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#number.
    def visitNumber(self, ctx: NeuralogicParser.NumberContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#vector.
    def visitVector(self, ctx: NeuralogicParser.VectorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#matrix.
    def visitMatrix(self, ctx: NeuralogicParser.MatrixContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#dimensions.
    def visitDimensions(self, ctx: NeuralogicParser.DimensionsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by NeuralogicParser#negation.
    def visitNegation(self, ctx: NeuralogicParser.NegationContext):
        return self.visitChildren(ctx)


del NeuralogicParser



from py4j.java_gateway import get_field, set_field
from typing import Optional, List, Iterable
from py4j.java_collections import ListConverter

from neuralogic import get_neuralogic, get_gateway
from neuralogic.core.settings import Settings


class JavaFactory:
    def __init__(self, settings: Optional[Settings] = None):
        from neuralogic.core.constructs.rule import Rule
        from neuralogic.core.constructs.atom import WeightedAtom

        if settings is None:
            settings = Settings()

        self.weighted_atom_type = WeightedAtom
        self.rule_type = Rule

        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.building
        self.settings = settings

        self.namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.template.components
        self.value_namespace = get_neuralogic().cz.cvut.fel.ida.algebra.values
        self.example_namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.example

        self.builder = namespace.TemplateBuilder(settings.settings)

        self.constant_factory = get_field(self.builder, "constantFactory")
        self.predicate_factory = get_field(self.builder, "predicateFactory")
        self.weight_factory = get_field(self.builder, "weightFactory")

        self.unit_weight = get_neuralogic().cz.cvut.fel.ida.algebra.weights.Weight.unitWeight

    def get_variable_factory(self):
        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.building.factories
        variable_factory = namespace.VariableFactory()

        return variable_factory

    def get_term(self, term, variable_factory):
        if isinstance(term, str):
            if term[0].islower() or term.isnumeric():
                return self.constant_factory.construct(term)
            elif term[0].isupper():
                return variable_factory.construct(term)
            else:
                raise NotImplementedError
        if isinstance(term, (int, float)):
            return self.constant_factory.construct(str(term))
        raise NotImplementedError

    def get_generic_atom(self, atom_class, atom, variable_factory, new_weight):
        predicate = self.get_predicate(atom.predicate)

        weight = None
        if isinstance(atom, self.weighted_atom_type):
            if new_weight:
                weight = self.get_weight(atom.weight, atom.is_fixed)
            else:
                weight = get_field(atom.java_object, "weight")

        term_list = ListConverter().convert(
            [self.get_term(term, variable_factory) for term in atom.terms], get_gateway()._gateway_client
        )

        java_atom = atom_class(predicate, term_list, atom.negated, weight)
        set_field(java_atom, "originalString", atom.to_str())

        return java_atom

    def get_metadata(self, metadata):
        if metadata is None:
            return None

        if (
            metadata.aggregation is None
            and metadata.activation is None
            and metadata.offset is None
            and metadata.learnable is None
        ):
            return None

        map = get_gateway().jvm.java.util.LinkedHashMap()

        if metadata.aggregation is not None:
            map.put("aggregation", self.value_namespace.StringValue(metadata.aggregation.value))
        if metadata.activation is not None:
            map.put("activation", self.value_namespace.StringValue(metadata.activation.value))
        # if metadata.offset is not None:
        #     _, value = self.get_value(metadata.offset)
        #     map.put("offset", self.weight_factory.construct(value))
        if metadata.learnable is not None:
            map.put("learnable", self.value_namespace.StringValue(str(metadata.learnable).lower()))

        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs.template.metadata
        return namespace.RuleMetadata(get_field(self.builder, "settings"), map)

    def get_query(self, query):
        if not isinstance(query, self.rule_type):
            if not isinstance(query, Iterable):
                query = [query]
            return None, self.get_conjunction(query)
        return query.head.java_object, self.get_conjunction(query.body)

    def get_lifted_example(self, example):
        gateway_client = get_gateway()._gateway_client

        conjunctions = []
        rules = ListConverter().convert([], gateway_client)
        label_conjunction = None

        if not isinstance(example, self.rule_type):
            if not isinstance(example, Iterable):
                example = [example]
            conjunctions.append(self.get_conjunction(example))
        else:
            label_conjunction = self.get_conjunction([example.head])
            conjunctions.append(self.get_conjunction(example.body))

        lifted_example = self.example_namespace.LiftedExample(
            ListConverter().convert(conjunctions, gateway_client), rules
        )
        return label_conjunction, lifted_example

    def get_conjunction(self, atoms):
        namespace = get_neuralogic().cz.cvut.fel.ida.logic.constructs
        valued_facts = [atom.java_object for atom in atoms]

        return namespace.Conjunction(ListConverter().convert(valued_facts, get_gateway()._gateway_client))

    def get_predicate_metadata_pair(self, predicate_metadata):
        namespace = get_neuralogic().cz.cvut.fel.ida.utils.generic
        return namespace.Pair(predicate_metadata.predicate.java_object, self.get_metadata(predicate_metadata.metadata))

    def get_valued_fact(self, atom, variable_factory):
        return self.get_generic_atom(self.example_namespace.ValuedFact, atom, variable_factory, True)

    def get_atom(self, atom, variable_factory):
        return self.get_generic_atom(self.namespace.BodyAtom, atom, variable_factory, False)

    def get_rule(self, rule):
        java_rule = self.namespace.WeightedRule()
        java_rule.setOriginalString(str(rule))

        variable_factory = self.get_variable_factory()

        head_atom = self.get_atom(rule.head, variable_factory)
        weight = head_atom.getConjunctWeight()

        if weight is None:
            java_rule.setWeight(self.unit_weight)
        else:
            java_rule.setWeight(weight)

        body_atoms = [self.get_atom(atom, variable_factory) for atom in rule.body]
        body_atom_list = ListConverter().convert(body_atoms, get_gateway()._gateway_client)

        java_rule.setHead(self.namespace.HeadAtom(head_atom))
        java_rule.setBody(body_atom_list)

        offset = None  # TODO: Implement

        java_rule.setOffset(offset)
        java_rule.setMetadata(self.get_metadata(rule.metadata))

        return java_rule

    def get_predicate(self, predicate):
        return self.predicate_factory.construct(predicate.name, predicate.arity, predicate.special, predicate.private)

    def get_weight(self, weight, fixed):
        initialized, value = self.get_value(weight)
        return self.weight_factory.construct(value, fixed, initialized)

    def get_value(self, weight):
        if isinstance(weight, (int, float)):
            value = self.value_namespace.ScalarValue(float(weight))
            initialized = True
        elif isinstance(weight, tuple):
            if len(weight) == 1:
                if weight[0] == 1:
                    value = self.value_namespace.ScalarValue()
                else:
                    value = self.value_namespace.VectorValue(weight[0])
            elif len(weight) == 2:
                if weight[0] == 1:
                    value = self.value_namespace.VectorValue(weight[1])
                    set_field(value, "rowOrientation", True)
                elif weight[1] == 1:
                    value = self.value_namespace.VectorValue(weight[0])
                    set_field(value, "rowOrientation", False)
                else:
                    value = self.value_namespace.MatrixValue(weight[0], weight[1])
            else:
                raise NotImplementedError
            initialized = False
        elif isinstance(weight, List):
            initialized = True
            if len(weight) == 0:
                raise NotImplementedError
            if isinstance(weight[0], (int, float)):
                vector = ListConverter().convert([float(w) for w in weight], get_gateway()._gateway_client)
                value = self.value_namespace.VectorValue(vector)
            elif isinstance(weight[0], List):
                matrix = []

                for weights in weight:
                    values = [float(w) for w in weights]
                    matrix.append(ListConverter().convert(values, get_gateway()._gateway_client))

                matrix = ListConverter().convert(matrix, get_gateway()._gateway_client)
                value = self.value_namespace.MatrixValue(matrix)
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError
        return initialized, value


java_factory: Optional[JavaFactory] = None


def set_java_factory(factory: Optional[JavaFactory]):
    global java_factory
    java_factory = factory


def get_java_factory() -> JavaFactory:
    if java_factory is None:
        raise Exception
    return java_factory


def get_current_java_factory() -> Optional[JavaFactory]:
    return java_factory



from typing import List, Optional, Dict
from neuralogic.core.builder import Sample, Weight, Neuron
import dynet as dy
import numpy as np

from neuralogic.core.settings import Settings


class NeuraLogicLayer:
    activations = {
        "Sigmoid": dy.logistic,
        "Average": dy.average,
        "Maximum": dy.emax,
        "ReLu": dy.rectify,
        "Tanh": dy.tanh,
    }

    def __init__(self, model: List[Weight], settings: Optional[Settings] = None):
        if settings is None:
            settings = Settings()
        self.settings = settings

        self.model = dy.ParameterCollection()
        self.weights_meta = model
        self.weights: List[dy.Parameters] = []

        self.reset_parameters()

    def reset_parameters(self):
        self.weights = [
            weight.value if weight.fixed else self.model.add_parameters(weight.dimensions, init="uniform")
            for weight in self.weights_meta
        ]

    def build_sample(self, sample: Sample) -> dy.Expression:
        dynet_neurons: List[Optional[dy.Expression]] = [None] * len(sample.neurons)

        for neuron in sample.neurons:
            dynet_neurons[neuron.index] = NeuraLogicLayer.to_dynet_expression(neuron, dynet_neurons, self.weights)
        return dynet_neurons[sample.neurons[-1].index]

    def __call__(self, sample: Sample) -> dy.Expression:
        return self.build_sample(sample)

    def state_dict(self) -> Dict:
        weights = {}

        for meta, weight in zip(self.weights_meta, self.weights):
            if not meta.fixed and meta.index >= 0:
                weights[meta.index] = weight if isinstance(weight, (int, float)) else weight.value()
        return {"weights": weights}

    def load_state_dict(self, state_dict: Dict):
        weight_dict = state_dict["weights"]

        for i, (meta, weight) in enumerate(zip(self.weights_meta, self.weights)):
            if meta.index < 0 or meta.fixed:
                continue

            if isinstance(weight, (int, float)):
                self.weights[i] = weight_dict[meta.index]
            else:
                weight.set_value(np.array(weight_dict[meta.index]).reshape(weight.shape()))

    @staticmethod
    def to_dynet_value(value) -> dy.Expression:
        dim = 1
        if hasattr(value, "__len__"):
            dim = len(value)
        if dim == 1:
            return dy.scalarInput(float(value))
        if dim > 1:
            res = dy.vecInput(dim)
            res.set(value)
            return res
        return dy.inputTensor(value, dim)

    @staticmethod
    def to_dynet_expression(neuron: Neuron, neurons: List[dy.Expression], weights: List):
        if neuron.inputs:
            out = NeuraLogicLayer.process_neuron_inputs(neuron, neurons, weights)
        else:
            out = NeuraLogicLayer.to_dynet_value(neuron.value)

        if neuron.activation:
            if neuron.pooling:
                out = list(out)
            out = NeuraLogicLayer.activations[neuron.activation](out)
        return out

    @staticmethod
    def process_neuron_inputs(neuron: Neuron, neurons: List, weights: List[dy.Parameters]) -> dy.Expression:
        if neuron.weights:
            if neurons[neuron.inputs[0]].dim()[0] == (1,) and not isinstance(weights[neuron.weights[0]], int):
                out = dy.cmult(neurons[neuron.inputs[0]], weights[neuron.weights[0]])
            else:
                out = weights[neuron.weights[0]] * neurons[neuron.inputs[0]]
            for w, i in zip(neuron.weights[1:], neuron.inputs[1:]):
                if neurons[i].dim()[0] == (1,) and not isinstance(weights[w], int):
                    out += dy.cmult(weights[w], neurons[i])
                else:
                    out += weights[w] * neurons[i]
        else:
            out = neurons[neuron.inputs[0]]
            for i in neuron.inputs[1:]:
                out += neurons[i]
        return out



from neuralogic.core.constructs.factories import Var, Term, Atom
from neuralogic.core.problem import Problem
from neuralogic.core.constructs.java_objects import JavaFactory, get_java_factory, set_java_factory
from neuralogic.core.constructs.metadata import Metadata, Activation, Aggregation
from neuralogic.core.builder import Backend
